@@ -25,25 +25,27 @@
 # ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 # POSSIBILITY OF SUCH DAMAGE.
 
-from ..util import parse_certificates
-from yubikit.core import TRANSPORT
-from yubikit.core.smartcard import SmartCardConnection, ApduError
-from yubikit.core.smartcard.scp import ScpKid, KeyRef, ScpKeyParams, Scp11KeyParams
-from yubikit.management import DeviceInfo, CAPABILITY
-from yubikit.oath import parse_b32_key
-from yubikit.securitydomain import SecurityDomainSession
+import functools
+import logging
+import sys
 from collections import OrderedDict
 from collections.abc import MutableMapping
-from cryptography.hazmat.primitives import serialization
-from cryptography import x509
-from contextlib import contextmanager
-from threading import Timer
 from enum import Enum
-from typing import Optional, Sequence, Tuple, List
-import functools
+from typing import Sequence, cast
+
 import click
-import sys
-import logging
+from cryptography import x509
+from cryptography.hazmat.primitives import serialization
+from cryptography.hazmat.primitives.asymmetric.ec import EllipticCurvePublicKey
+
+from yubikit.core import TRANSPORT, ApplicationNotAvailableError, _timeout
+from yubikit.core.smartcard import ApduError, SmartCardConnection
+from yubikit.core.smartcard.scp import KeyRef, Scp11KeyParams, ScpKeyParams, ScpKid
+from yubikit.management import CAPABILITY, DeviceInfo
+from yubikit.oath import parse_b32_key
+from yubikit.securitydomain import SecurityDomainSession
+
+from ..util import parse_certificates
 
 logger = logging.getLogger(__name__)
 
@@ -63,6 +65,7 @@ class _YkmanCommand(click.Command):
 
     def get_help_option(self, ctx):
         option = super().get_help_option(ctx)
+        assert option is not None  # noqa: S101
         option.help = "show this message and exit"
         return option
 
@@ -266,22 +269,16 @@ def click_prompt(prompt, err=True, **kwargs):
     return click.prompt(prompt, err=err, **kwargs)
 
 
-def prompt_for_touch():
+def prompt_for_touch(prompt: str = "Touch your YubiKey...") -> None:
     logger.debug("Prompting user to touch YubiKey...")
     try:
-        click.echo("Touch your YubiKey...", err=True)
+        click.echo(prompt, err=True)
     except Exception:
-        sys.stderr.write("Touch your YubiKey...\n")
+        sys.stderr.write(f"{prompt}\n")
 
 
-@contextmanager
 def prompt_timeout(timeout=0.5):
-    timer = Timer(timeout, prompt_for_touch)
-    try:
-        timer.start()
-        yield None
-    finally:
-        timer.cancel()
+    return _timeout(prompt_for_touch, timeout)
 
 
 class CliFail(Exception):
@@ -296,7 +293,7 @@ def pretty_print(value, level: int = 0) -> Sequence[str]:
     Returns a list of strings which can be printed as lines.
     """
     indent = "  " * level
-    lines: List[str] = []
+    lines: list[str] = []
     if isinstance(value, list):
         for v in value:
             lines.extend(pretty_print(v, level))
@@ -348,17 +345,34 @@ def log_or_echo(message: str, log: logging.Logger, *files) -> None:
 
 
 def find_scp11_params(
-    connection: SmartCardConnection, kid: int, kvn: int, ca: Optional[bytes] = None
+    connection: SmartCardConnection, kid: int, kvn: int, ca: bytes | None = None
 ) -> Scp11KeyParams:
-    scp = SecurityDomainSession(connection)
+    try:
+        scp = SecurityDomainSession(connection)
+    except ApplicationNotAvailableError:
+        raise ValueError("Security Domain application not available")
+
+    if ca:
+        root_ca = parse_certificates(ca, None)[0]
+        try:
+            ski = root_ca.extensions.get_extension_for_class(x509.SubjectKeyIdentifier)
+        except x509.ExtensionNotFound:
+            ski = None
+    else:
+        root_ca = None
+        ski = None
+
     if not kvn:
-        if ca:
+        if ski:
             # Find by CA
+            ca_ski = ski.value.digest
             for ref, ca_check in scp.get_supported_ca_identifiers(klcc=True).items():
-                if ca_check == ca:
+                if ca_check == ca_ski:
                     if not kid or ref.kid == kid:
                         kid, kvn = ref
                         break
+            else:
+                raise ValueError(f"No CA identifier found matching SKI: {ca_ski.hex()}")
         # Find any matching KID
         for ref in scp.get_key_information().keys():
             if ref.kid == kid:
@@ -366,14 +380,15 @@ def find_scp11_params(
                 break
         else:
             raise ValueError(f"No SCP key found matching KID=0x{kid:x}")
+
+    ref = KeyRef(kid, kvn)
     try:
-        ref = KeyRef(kid, kvn)
         chain = scp.get_certificate_bundle(ref)
         if not chain:
             raise ValueError(f"No certificate chain stored for {ref}")
-        if ca:
+        if root_ca:
             logger.debug("Validating KLCC CA using supplied file")
-            parent = parse_certificates(ca, None)[0]
+            parent = root_ca
             for cert in chain:
                 # Requires cryptography >= 40
                 cert.verify_directly_issued_by(parent)
@@ -382,7 +397,7 @@ def find_scp11_params(
         else:
             logger.info("No CA supplied, skipping KLCC CA validation")
 
-        pub_key = chain[-1].public_key()
+        pub_key = cast(EllipticCurvePublicKey, chain[-1].public_key())
         return Scp11KeyParams(ref, pub_key)
     except ApduError:
         raise ValueError(f"Unable to get SCP key paramaters ({ref})")
@@ -390,7 +405,7 @@ def find_scp11_params(
 
 def get_scp_params(
     ctx: click.Context, capability: CAPABILITY, connection: SmartCardConnection
-) -> Optional[ScpKeyParams]:
+) -> ScpKeyParams | None:
     # Explicit SCP
     resolve = ctx.obj.get("scp")
     if resolve:
@@ -412,8 +427,8 @@ def get_scp_params(
 
 def organize_scp11_certificates(
     certificates: Sequence[x509.Certificate],
-) -> Tuple[
-    Optional[x509.Certificate], Sequence[x509.Certificate], Optional[x509.Certificate]
+) -> tuple[
+    x509.Certificate | None, Sequence[x509.Certificate], x509.Certificate | None
 ]:
     if not certificates:
         return None, [], None
@@ -442,7 +457,7 @@ def organize_scp11_certificates(
 
     # Check if leaf has keyAgreement policy:
     if ordered:
-        kue = ordered[-1].extensions.get_extension_for_oid(x509.ExtensionOID.KEY_USAGE)
+        kue = ordered[-1].extensions.get_extension_for_class(x509.KeyUsage)
         if kue.value.key_agreement:
             leaf = ordered.pop()
 
